import 'package:collection/collection.dart';
import 'package:meta/meta.dart';
import 'package:recase/recase.dart' as recase;

import 'case.dart';
import 'category.dart';
import 'condition.dart';
import 'template_value.dart';

/// Represents a label that can have multiple translations.
class Label {
  Label({
    @required this.key,
    @required this.cases,
  }) : assert(_areCasesValid(key, cases));

  final String key;
  String get normalizedKey => recase.ReCase(key).camelCase;
  final List<Case> cases;
  List<TemplatedValue> get templatedValues {
    if (cases.isNotEmpty) {
      final templatedValues = cases.first.templatedValues;
      for (var i = 1; i < cases.length; i++) {
        final current = cases[i];
        assert(
            const SetEquality().equals(
                templatedValues.toSet(), current.templatedValues.toSet()),
            'All cases should have the same template values');
      }

      return templatedValues;
    }

    return [];
  }

  void addCase(Case newCase) {
    cases.add(newCase);
    _areCasesValid(key, cases);
  }

  Category get category {
    final values = cases
        .where((x) => x.condition is CategoryCondition)
        .map((x) => x.condition as CategoryCondition);

    if (values.isNotEmpty) {
      return Category(values.first.category.name)
        ..values.addAll(values.map((x) => x.value));
    }

    return null;
  }

  static bool _areCasesValid(String key, List<Case> cases) {
    final defaultCases =
        cases.where((x) => x.condition is DefaultCondition).length;

    assert(defaultCases <= 1,
        'There is more than one default case for label with key `$key`');

    final categories = cases
        .where((x) => x.condition is CategoryCondition)
        .map((x) => (x.condition as CategoryCondition).category?.name)
        .where((x) => x != null)
        .toSet();

    assert(categories.length <= 1,
        'There is more than one category in conditions for label `$key`');

    return true;
  }
}

/// Represents a translation of a label in a given language.
class Translation {
  final Case condition;
  final String languageCode;
  final String value;
  final List<TemplatedValue> templatedValues;

  Translation(
    this.languageCode,
    this.value, {
    this.condition,
  }) : templatedValues = TemplatedValue.parse(value);
}
